/*
 * e4graph.h --
 *
 *	This file contains the definition of the public interface part of the
 *	e4Graph library. It defines the following classes:
 *
 *	e4_RefCount	-- A base class used to manage lifecycle.
 *	e4_Storage	-- A representation of tree data storage.
 *	e4_Marker	-- A representation of a marker.
 *	e4_Node		-- A representation of a node.
 *	e4_Vertex	-- A representation of a vertex.
 *	e4_StorageVisitor --
 *			   A visitor ranging over the currently open
 *			   storages.
 *	e4_VertexVisitor-- A visitor ranging over various selections of
 *			   vertices in a storage.
 *	e4_MarkerVisitor-- A visitor ranging over the markers in a storage.
 *	e4_NodeVisitor	-- A visitor ranging over the nodes in a storage.
 *
 * Copyright (c) 2000-2001, JYL Software Inc.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE, EVEN IF
 * JYL SOFTWARE INC. IS MADE AWARE OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef	__E4_GRAPH_H__
#define	__E4_GRAPH_H__

#include "e4common.h"

/*
 * Forward declaration of e4_Value.
 */

struct e4_Value;

/*
 * The e4_RefCount class:
 */

class e4_DLL e4_RefCount {
protected:

    /*
     * The "impl" field contains the reference to the actual referenced
     * implementation instance.
     */

    e4_RefCounter *impl;

public:
    /*
     * Default constructor:
     */

    e4_RefCount();

    /*
     * Copying constructor:
     */

    e4_RefCount(const e4_RefCount &referrer);

    /*
     * Assignment operator:
     */

    e4_RefCount & operator=(const e4_RefCount &referrer);

    /*
     * Destructor:
     */

    ~e4_RefCount();

    /*
     * Returns the current refcount if the impl field is valid, or
     * -1 if not.
     */

    int RefCount() const;

    /*
     * Is this instance valid?
     */

    bool IsValid() const;

    /*
     * Return an identifier for programmatic type identification for
     * this instance.
     */

    virtual e4_RefKind Kind() const;

    /*
     * Comparison operators.
     */

    bool operator==(const e4_RefCount &compared) const;
    bool operator!=(const e4_RefCount &compared) const;

    /*
     * This operation returns an integer that is guaranteed to be unique
     * while "impl" is valid. It may be different each time the underlying
     * "impl" is instantiated.
     */

    inline int GetTemporaryUID() {
	return (int) impl;
    }

    /*
     * These operations manipulate user-level data attached to each instance.
     */

    void SetTransientUserData(void *data) const;
    void *GetTransientUserData() const;

protected:

    /*
     * Constructor with value assignd to implementation.
     */

    e4_RefCount(e4_RefCounter *ip);
};

/*
 * The e4_Storage class:
 */

class e4_DLL e4_Storage : public e4_RefCount {
public:

    /*
     * Default constructor:
     */

    e4_Storage();

    /*
     * Copying constructor:
     */

    e4_Storage(const e4_Storage &referrer);

    /*
     * Assignment operator:
     */

    e4_Storage & operator=(const e4_Storage &referrer);

    /*
     * No destructor is defined so the e4_RefCount destructor will be used.
     */

    /*
     * Constructor that chooses a concrete representation for the
     * underlying implementation.
     */

    e4_Storage(const char *name, const char *drivername);

    /*
     * Turn auto-commit on or off.
     */

    void AutoCommit(bool on) const;

    /*
     * Is auto-commit turned on?
     */

    bool AutoCommit() const;

    /*
     * Commit now.
     */

    bool Commit() const;

    /*
     * Copy the content (e4Graph related only) of this storage to another
     * storage.
     */

    bool CopyTo(e4_Storage otherStorage, bool forceCommit) const;

    /*
     * Delete the underlying physical storage.
     */

    bool Delete();

    /*
     * How many markers are there in this storage?
     */

    int MarkerCount() const;

    /*
     * Return a marker from the storage.
     */

    bool GetMarker(const char *nm, e4_Marker &t) const;

    /*
     * Given a unique ID, obtain the marker, node or vertex denoted
     * by that ID.
     */

    bool GetMarkerFromID(e4_MarkerUniqueID mid, e4_Marker &t) const;
    bool GetNodeFromID(e4_NodeUniqueID nid, e4_Node &n) const;
    bool GetVertexFromID(e4_VertexUniqueID id, e4_Vertex &v) const;

    /*
     * Does this storage contain a marker with the given name?
     */

    bool MarkerExists(const char *nm);

    /*
     * Retrieve the node marked by the named marker. The marker
     * is created if it does not exist.
     */

    bool GetMarkedNode(const char *nm, e4_Node &n) const;

    /*
     * Make this storage empty.
     */

    bool MakeEmpty() const;

    /*
     * Is this storage empty?
     */

    bool IsEmpty() const;

    /*
     * Is this storage dirty?
     */

    bool IsDirty() const;

    /*
     * Is the storage locked (temporarily) while a deletion is in
     * progress?
     */

    bool IsLocked() const;

    /*
     * Mark the storage as dirty (it needs to be committed).
     */

    void MarkDirty() const;

    /*
     * Get the name and driver ID for this storage.
     */

    const char *GetName() const;
    const char *GetDriver() const;

    /*
     * Get a value for a statistic being collected for this storage.
     */

    bool GetStatistic(e4_Space sp, e4_SpaceStat st, int &v) const;

    /*
     * Callback management: add or delete callbacks.
     */

    bool DeclareMarkerAddCallback(e4_MarkerAddCallbackFunction fn,
				  void *clientData);
    bool DeleteMarkerAddCallback(e4_MarkerAddCallbackFunction fn,
				 void *clientData);
    bool DeclareMarkerDelCallback(e4_MarkerDelCallbackFunction fn,
				  void *clientData);
    bool DeleteMarkerDelCallback(e4_MarkerDelCallbackFunction fn,
				 void *clientData);
    bool DeclareNodeAddCallback(e4_NodeAddCallbackFunction fn,
				void *clientData);
    bool DeleteNodeAddCallback(e4_NodeAddCallbackFunction fn,
			       void *clientData);
    bool DeclareNodeDelCallback(e4_NodeDelCallbackFunction fn,
				void *clientData);
    bool DeleteNodeDelCallback(e4_NodeDelCallbackFunction fn,
			       void *clientData);
    bool DeclareVertexAddCallback(e4_VertexAddCallbackFunction fn,
				  void *clientData);
    bool DeleteVertexAddCallback(e4_VertexAddCallbackFunction fn,
				 void *clientData);
    bool DeclareVertexDelCallback(e4_VertexDelCallbackFunction fn,
				  void *clientData);
    bool DeleteVertexDelCallback(e4_VertexDelCallbackFunction fn,
				 void *clientData);
    bool DeclareVertexModCallback(e4_VertexModCallbackFunction fn,
				  void *clientData);
    bool DeleteVertexModCallback(e4_VertexModCallbackFunction fn,
				 void *clientData);

#ifdef	NOTDEF
    /*
     * Does this storage need a GC?
     */

    bool NeedsGC() const;

    /*
     * Do a GC on this storage.
     */

    bool DoGC() const;
#endif

    /*
     * Get the kind identifier for an instance of this type:
     */

    virtual e4_RefKind Kind() const;

protected:

    /*
     * These methods are for use by classes e4_Marker, e4_Node,
     * e4_Vertex, e4_StorageVisitor, e4_VertexVisitor, e4_MarkerVisitor,
     * e4_NodeVisitor, e4_MarkerUniqueID, e4_NodeUniqueID, and
     * e4_VertexUniqueID only.
     */

    friend class e4_Marker;
    friend class e4_Node;
    friend class e4_Vertex;
    friend class e4_StorageVisitor;
    friend class e4_VertexVisitor;
    friend class e4_MarkerVisitor;
    friend class e4_NodeVisitor;

    /*
     * Constructor which assigns a value to the implementation field:
     */

    e4_Storage(e4_StorageImpl *ip);

    /*
     * Get the next storage after this one in an implementation dependent
     * order that visits all currently open storages.
     */

    bool FindNextStorage(e4_Storage &ss) const;

    /*
     * Get the next vertex after a vertex identified by a unique ID,
     * using a search method specified by "vm" and "vf". These are
     * defined in e4common.h.
     */

    bool FindNextVertex(int vertexID, e4_VisitMethod vm, int vf,
			int nameId, int nodeID,
			e4_VertexType typeID, e4_Vertex &f) const;

    /*
     * Find the next marker after a marker identified by a unique ID.
     */

    bool FindNextMarker(int markerID, e4_Marker &t) const;

    /*
     * Find the next node after a node identified by a unique ID.
     */

    bool FindNextNode(int nodeID, e4_Node &n) const;

    /*
     * Intern a name.
     */

    int InternName(const char *nm) const;

    /*
     * Get a marker given its marker ID:
     */

    e4_MarkerImpl *GetMarker(int markerID) const;

    /*
     * Get a node given its node ID:
     */

    e4_NodeImpl *GetNode(int nodeID) const;

    /*
     * Get the name of a vertex given the name ID.
     */

    const char *GetName(int nameID) const;

    /*
     * Get the storage implementation from this storage:
     */

    e4_StorageImpl *GetStorageImpl() const;
};

/*
 * The e4_Marker class:
 */

class e4_DLL e4_Marker : public e4_RefCount {
public:

    /*
     * Default constructor:
     */

    e4_Marker();

    /*
     * Copying constructor:
     */

    e4_Marker(const e4_Marker &referrer);

    /*
     * Assignment operator:
     */

    e4_Marker & operator=(const e4_Marker &referrer);

    /*
     * No destructor is defined so the e4_RefCount destructor will be used.
     */

    /*
     * Get the name of this marker.
     */

    const char *GetName() const;

    /*
     * Get the storage containing this marker.
     */

    bool GetStorage(e4_Storage &s) const;

    /*
     * Get the node marked by this marker.
     */

    bool GetMarkedNode(e4_Node &n) const;

    /*
     * Move this marker to another node.
     */

    bool SetMarkedNode(e4_Node n) const;

    /*
     * Get a unique identifier for this marker (unique within its storage).
     */

    bool GetUniqueID(e4_MarkerUniqueID &m) const;

    /*
     * Delete the storage occuppied by this marker.
     */

    bool Delete() const;

    /*
     * Return the kind identifier for an instance of this type:
     */

    virtual e4_RefKind Kind() const;

    /*
     * Get/Set user data associated with this marker.
     */

    bool GetUserData(int &userData) const;
    bool SetUserData(int userData) const;

protected:

    /*
     * These operations are for use by classes e4_Storage, e4_Node,
     * e4_Vertex, e4_StorageVisitor, e4_VertexVisitor, e4_MarkerVisitor,
     * and e4_NodeVisitor only:
     */

    friend class e4_Storage;
    friend class e4_Node;
    friend class e4_Vertex;
    friend class e4_StorageVisitor;
    friend class e4_VertexVisitor;
    friend class e4_MarkerVisitor;
    friend class e4_NodeVisitor;
    friend class e4_StorageImpl;

    /*
     * Constructor which assigns a value to the implementation field:
     */

    e4_Marker(e4_MarkerImpl *ip);

    /*
     * Get the unique ID for this marker in a raw (unwrapped) form.
     */

    int GetRawUniqueID() const;
};

/*
 * The e4_Node class:
 */

class e4_DLL e4_Node : public e4_RefCount {
public:

    /*
     * Default constructor:
     */

    e4_Node();

    /*
     * Copying constructor:
     */

    e4_Node(const e4_Node &referrer);

    /*
     * Assignment operator:
     */

    e4_Node & operator=(const e4_Node &referrer);

    /*
     * No destructor is defined so the e4_RefCount destructor will be used.
     */

    /*
     * How many vertices are there in this node?
     */

    int VertexCount() const;

    /*
     * Operations to set data into existing vertices in this node.
     */

    bool SetNthVertex(const char *nm, int nth, int i) const;
    bool SetNthVertex(const char *nm, int nth, double f) const;
    bool SetNthVertex(const char *nm, int nth, const char *s) const;
    bool SetNthVertex(const char *nm, int nth, const void *b, int nb) const;
    bool SetNthVertex(const char *nm, int nth, e4_Node n) const;
    bool SetNthNode(const char *nm, int nth, e4_Node &n) const;

    inline bool SetVertex(const char *nm, int i) const {
        return SetNthVertex(nm, 1, i);
    }
    inline bool SetVertex(const char *nm, double f) const {
        return SetNthVertex(nm, 1, f);
    }
    inline bool SetVertex(const char *nm, const char *s) const {
        return SetNthVertex(nm, 1, s);
    }
    inline bool SetVertex(const char *nm, const void *b, int nb) const {
        return SetNthVertex(nm, 1, b, nb);
    }
    inline bool SetVertex(const char *nm, e4_Node n) const {
	return SetNthVertex(nm, 1, n);
    }
    inline bool SetNode(const char *nm, e4_Node &n) const {
        return SetNthNode(nm, 1, n);
    }

    /*
     * Set the value of an existing vertex in this node identified by rank.
     */

    bool SetVertexByRank(int rank, int i) const;
    bool SetVertexByRank(int rank, double f) const;
    bool SetVertexByRank(int rank, const char *s) const;
    bool SetVertexByRank(int rank, const void *b, int nb) const;
    bool SetVertexByRank(int rank, e4_Node n) const;
    bool SetNodeByRank(int rank, e4_Node &n) const;

    /*
     * Add a new vertex to this node according to the rank and order arguments.
     * The rank argument is in-out; on the way in, in combination with the
     * order argument, it determines where in the node the new vertex is
     * created. On the way out, it returns the rank of the newly created
     * vertex. Upon success, these operations return true. If an error occurred
     * these operations return false.
     */

    bool AddVertex(const char *nm, e4_InsertOrder order, int &rank,
		   int value) const;
    bool AddVertex(const char *nm, e4_InsertOrder order, int &rank,
		   double f) const;
    bool AddVertex(const char *nm, e4_InsertOrder order, int &rank,
		   const char *s) const;
    bool AddVertex(const char *nm, e4_InsertOrder order, int &rank,
		   const void *b, int nb) const;
    bool AddVertex(const char *nm, e4_InsertOrder order, int &rank,
		   e4_Node n) const;
    bool AddNode(const char *nm, e4_InsertOrder order, int &rank, 
		 e4_Node &n) const;

    /*
     * Add a new vertex and also return an e4_Vertex for the new vertex.
     * This is much more efficient than AddVertex followed by GetVertexRef.
     */

    bool AddVertexRef(const char *nm, e4_InsertOrder order, int &rank,
                     int ii, e4_Vertex &f) const;
    bool AddVertexRef(const char *nm, e4_InsertOrder order, int &rank,
                     double ff, e4_Vertex &f) const;
    bool AddVertexRef(const char *nm, e4_InsertOrder order, int &rank,
                     const char *ss, e4_Vertex &f) const;
    bool AddVertexRef(const char *nm, e4_InsertOrder order, int &rank,
                     const void *b, int nb, e4_Vertex &f) const;
    bool AddVertexRef(const char *nm, e4_InsertOrder order, int &rank,
		     e4_Node n, e4_Vertex &f) const;
    bool AddNodeRef(const char *nm, e4_InsertOrder order, int &rank,
                    e4_Node &n, e4_Vertex &f) const;

    /*
     * Move a given vertex to a specified position within this node.
     */

    bool MoveVertex(e4_Vertex &f, e4_InsertOrder order, int rank);

    /*
     * Get a value out of a vertex. If the vertex is found then the output
     * parameter is set, otherwise it is not modified.
     *
     * The char * and void * returned are valid only until the next call to
     * any e4Graph API. If the caller needs to keep access to the value,
     * copy the value into memory owned by the caller.
     */
    
    bool GetNthVertex(const char *nm, int nth, e4_Value &v) const;
    bool GetNthVertex(const char *nm, int nth, e4_Node &n) const;
    bool GetNthVertex(const char *nm, int nth, int &v) const;
    bool GetNthVertex(const char *nm, int nth, double &v) const;
    bool GetNthVertex(const char *nm, int nth, const char *&v) const;
    bool GetNthVertex(const char *nm, int nth, const void *&v, int &nbv) const;

    inline bool GetVertex(const char *nm, e4_Value &v) const {
        return GetNthVertex(nm, 1, v);
    }
    inline bool GetVertex(const char *nm, e4_Node &v) const {
        return GetNthVertex(nm, 1, v);
    }
    inline bool GetVertex(const char *nm, int &v) const {
        return GetNthVertex(nm, 1, v);
    }
    inline bool GetVertex(const char *nm, double &v) const {
        return GetNthVertex(nm, 1, v);
    }
    inline bool GetVertex(const char *nm, const char *&s) const {
        return GetNthVertex(nm, 1, s);
    }
    inline bool GetVertex(const char *nm, const void *&v, int &nbv) const {
        return GetNthVertex(nm, 1, v, nbv);
    }

    /*
     * Get the value of a vertex identified by rank, in this node.
     */

    bool GetVertexByRank(int rank, e4_Value &v) const;
    bool GetVertexByRank(int rank, e4_Node &n) const;
    bool GetVertexByRank(int rank, int &v) const;
    bool GetVertexByRank(int rank, double &v) const;
    bool GetVertexByRank(int rank, const char *&s) const;
    bool GetVertexByRank(int rank, const void *&v, int &nbv) const;

    /*
     * Delete a vertex.
     */

    bool DeleteVertex(const char *nm, int nth) const;
    inline bool DeleteVertex(const char *nm) const {
        return DeleteVertex(nm, 1);
    }
    bool DeleteVertexByRank(int rank) const;

    /*
     * Delete the first vertex whose value is a given child node.
     */

    bool DeleteFirstVertexWithNode(e4_Node child);

    /*
     * Get an e4_Vertex for a vertex identified by name or rank.
     */

    bool GetVertexRef(const char *nm, int nth, e4_Vertex &f) const;
    inline bool GetVertexRef(const char *nm, e4_Vertex &f) const {
        return GetVertexRef(nm, 1, f);
    }
    bool GetVertexRefByRank(int rank, e4_Vertex &f) const;

    /*
     * Get the type of the value stored in a specific vertex.
     */

    e4_VertexType VertexType(const char *nm, int nth) const;
    inline e4_VertexType VertexType(const char *nm) const {
        return VertexType(nm, 1);
    }
    e4_VertexType VertexTypeByRank(int rank) const;

    /*
     * Get the name of a vertex.
     */

    const char *VertexName(int rank) const;

    /*
     * Rename a vertex.
     */

    bool RenameVertex(int rank, const char *newName) const;

    /*
     * Get the rank of a vertex.
     */

    int VertexRank(const char *nm, int nth) const;
    inline int VertexRank(const char *nm) const{
        return VertexRank(nm, 1);
    }

    /*
     * Does a named vertex exist?
     */

    bool Exists(const char *nm, int nth) const;
    inline bool Exists(const char *nm) const{
        return Exists(nm, 1);
    }

    /*
     * Get a parent node of a node.
     */

    bool GetParent(int nth, e4_Node &p) const;
    inline bool GetParent(e4_Node &p) const {
	return GetParent(1, p);
    }

    /*
     * Is this node marked?
     */

    bool IsMarked() const;

    /*
     * Retrieve the nth marker that marks this node.
     */

    bool GetMarker(int nth, e4_Marker &m) const;
    inline bool GetMarker(e4_Marker &m) const {
	return GetMarker(1, m);
    }

    /*
     * Add a new marker to this node.
     */

    bool AddMarker(const char *name, e4_Marker &m) const;

    /*
     * Make the given marker be a marker for this node.
     */

    bool MarkWith(e4_Marker m) const;

    /*
     * Make the given marker not mark this node. This also
     * removes the marker from the storage. If the node becomes
     * unreachable then it and all its unreachable descendants
     * are deleted from the storage.
     */

    bool Unmark(e4_Marker m) const;

    /*
     * Make the nth marker not mark this node. This also
     * removes the marker from the storage. If the node becomes
     * unreachable then it and all its unreachable descendants
     * are deleted from the storage.
     */

    bool Unmark(int nth) const;

    /*
     * How many markers mark this node?
     */

    int MarkerCount() const;

    /*
     * How many parents does this node have?
     */

    int ParentCount() const;

    /*
     * How many vertices contain this node as the value?
     */

    int OccurrenceCount() const;

    /*
     * How many times does this node occur as a child in the given node?
     */

    int OccurrenceCount(e4_Node p) const;

    /*
     * What is the rank of this parent for this node?
     */

    int ParentRank(e4_Node p) const;
    
    /*
     * Get the rank of the first vertex containing this node in its parent 
     * and the name of the first vertex containing this node in its parent.
     */

    int GetRankInParent(int nth) const;
    inline int GetRankInParent() const {
	return GetRankInParent(1);
    }

    /*
     * A node is a "root" if it has no parents. Invariantly, such a node
     * must be marked with at least one marker.
     */

    bool IsRoot() const;

    const char *GetNameInParent(int nth) const;
    inline const char *GetNameInParent() const {
	return GetNameInParent(1);
    }
    
    /*
     * Get the storage containing this node.
     */

    bool GetStorage(e4_Storage &s) const;

    /*
     * Get a unique ID (unique within the storage containing this node) for
     * the node.
     */

    bool GetUniqueID(e4_NodeUniqueID &n) const;
    
    /*
     * Delete the persistent storage occupied by this node.
     */

    bool Delete() const;

    /*
     * Return the kind identifier for an instance of this type:
     */

    virtual e4_RefKind Kind() const;

    /*
     * Get/Set user data associated with this node.
     */

    bool GetUserData(int &userData) const;
    bool SetUserData(int userData) const;

protected:

    /*
     * These operations are for use by class e4_Storage, e4_Marker, e4_Vertex,
     * e4_VertexVisitor, e4_MarkerVisitor, e4_StorageVisitor,
     * and e4_NodeVisitor only:
     */

    friend class e4_Storage;
    friend class e4_Marker;
    friend class e4_Vertex;
    friend class e4_StorageVisitor;
    friend class e4_VertexVisitor;
    friend class e4_MarkerVisitor;
    friend class e4_NodeVisitor;
    friend class e4_StorageImpl;

    /*
     * Constructor which assigns a value to the implementation field:
     */

    e4_Node(e4_NodeImpl *ip);

    /*
     * Make the marker identified by markerID not mark this node anymore.
     */

    bool UnmarkID(int markerID) const;

    /*
     * Get the unique ID for this node in a raw (unwrapped) form.
     */

    int GetRawUniqueID() const;
};

/*
 * The e4_Vertex class:
 */

class e4_DLL e4_Vertex : public e4_RefCount {
public:
    /*
     * Default constructor:
     */

    e4_Vertex();

    /*
     * Copying constructor:
     */

    e4_Vertex(const e4_Vertex &referrer);

    /*
     * Assignment operator:
     */

    e4_Vertex & operator=(const e4_Vertex &referrer);

    /*
     * No destructor is defined so the e4_RefCount destructor will be used.
     */

    /*
     * Get the value stored in this vertex.
     */

    bool Get(e4_Value &v) const;
    bool Get(e4_Node &n) const;
    bool Get(int &v) const;
    bool Get(double &v) const;
    bool Get(const char *&s) const;
    bool Get(const void *&b, int &nb) const;

    /*
     * Set the content of this vertex (and its type) to the value supplied.
     */

    bool Set(int v) const;
    bool Set(double v) const;
    bool Set(const char *v) const;
    bool Set(const void *b, int nb) const;
    bool Set(e4_Node n) const;

    /*
     * Set the content of this vertex to a new node and return the new node.
     */

    bool SetNode(e4_Node &n) const;

    /*
     * Get the rank of this vertex in the node in which it appears.
     */

    int Rank() const;

    /*
     * Delete the vertex from the node in which it appears.
     */

    bool Delete() const;

    /*
     * Return the type of the vertex.
     */

    e4_VertexType Type() const;

    /*
     * Return the name of the vertex.
     */

    const char *Name() const;

    /*
     * Rename this vertex.
     */

    bool Rename(const char *newname) const;

    /*
     * Get a unique (within the storage containing it) ID for this vertex.
     */

    bool GetUniqueID(e4_VertexUniqueID &v) const;

    /*
     * Get the storage in which this vertex appears.
     */

    bool GetStorage(e4_Storage &ss) const;

    /*
     * Get the node in which this vertex appears.
     */

    bool GetNode(e4_Node &nn) const;

    /*
     * Is the origin node of this vertex marked?
     */

    bool IsMarked() const;

    /*
     * Move the given vertex into the node containing this vertex at
     * the indicated insertOrder and offset. If insertOrder is either
     * E4_IOBEFORE, offset is subtracted from the rank of this vertex
     * plus one to get the rank at which the given vertex is inserted.
     * If insertOrder is E4_IOAFTER, offset is added to the rank of
     * this vertex to get the rank at which the given vertex is inserted.
     * In these cases, offset must be equal or greater than one.
     * Otherwise offset is ignored.
     */

    bool MoveVertex(e4_Vertex &ff, e4_InsertOrder order, int offset) const;

    /*
     * Retrieve the next vertex after this one in the node
     * containing this vertex.
     */

    bool Next(int num, e4_Vertex &ff) const;

    /*
     * Retrieve the previous vertex before this one in the node
     * containing this vertex.
     */

    bool Prev(int num, e4_Vertex &ff) const;

    /*
     * Return the kind identifier for an instance of this type:
     */

    virtual e4_RefKind Kind() const;

    /*
     * Get/Set user data associated with this vertex.
     */

    bool GetUserData(int &userData) const;
    bool SetUserData(int userData) const;

protected:

    /*
     * These operations are for use by class e4_Node, e4_Storage,
     * e4_VertexVisitor, e4_MarkerVisitor, e4_StorageVisitor,
     * and e4_NodeVisitor only:
     */

    friend class e4_Node;
    friend class e4_Storage;
    friend class e4_StorageVisitor;
    friend class e4_VertexVisitor;
    friend class e4_MarkerVisitor;
    friend class e4_NodeVisitor;
    friend class e4_StorageImpl;
    
    /*
     * Constructor which assigns a value to the implementation field:
     */

    e4_Vertex(e4_VertexImpl *ip);

    /*
     * Get the unique ID for this vertex in a raw (unwrapped) form.
     */

    int GetRawUniqueID() const;
};

/*
 * Class e4_StorageVisitor
 */

class e4_DLL e4_StorageVisitor {
private:

    /*
     * The storage being visited now.
     */

    e4_Storage s;

    /*
     * Is this visitor done visiting all the storages?
     */

    bool done;

public:

    /*
     * Default constructor:
     */

    e4_StorageVisitor();

    /*
     * Copying constructor:
     */

    e4_StorageVisitor(const e4_StorageVisitor &referrer);

    /*
     * Assignment operator:
     */

    e4_StorageVisitor & operator=(const e4_StorageVisitor &referrer);

    /*
     * Comparison operators:
     */

    bool operator==(const e4_StorageVisitor &compared) const;
    bool operator!=(const e4_StorageVisitor &compared) const;

    /*
     * Destructor:
     */

    ~e4_StorageVisitor();

    /*
     * Is this visitor done visiting all the storages?
     */

    bool IsDone();

    /*
     * Get the current storage being visited:
     */

    bool CurrentStorage(e4_Storage &ss);

    /*
     * Get the current storage being visited and advance the
     * visitor to the next storage.
     */

    bool CurrentStorageAndAdvance(e4_Storage &ss);

    /*
     * Get the next storage:
     */

    bool NextStorage(e4_Storage &ss);

    /*
     * Is this instance valid?
     */

    bool IsValid();
};

/*
 * Class e4_VertexVisitor:
 */

class e4_DLL e4_VertexVisitor {
private:

    /*
     * The storage used by this visitor.
     */

    e4_Storage s;

    /*
     * The last vertex returned by the visitor.
     */

    e4_Vertex f;

    /*
     * Is this visitor done visiting all vertices in its itinerary?
     */

    bool done;

    /*
     * The ID of the node whose vertices we will visit.
     */

    int nodeID;

    /*
     * Only vertices with this nameID will be visited.
     */

    int nameID;

    /*
     * Only vertices with this type will be visited.
     */

    e4_VertexType typeID;

    /*
     * The method by which vertices are visited.
     */

    e4_VisitMethod vm;

    /*
     * Vertices to be visited are chosen by this visit filter.
     */

    int vf;

public:

    /*
     * Default constructor:
     */

    e4_VertexVisitor();

    /*
     * Copying constructor:
     */

    e4_VertexVisitor(const e4_VertexVisitor &referrer);

    /*
     * Assignment operator:
     */

    e4_VertexVisitor & operator=(const e4_VertexVisitor &referrer);

    /*
     * Comparison operators:
     */

    bool operator==(const e4_VertexVisitor &compared) const;
    bool operator!=(const e4_VertexVisitor &compared) const;

    /*
     * Constructors which initialize the visitor from a node.
     */

    e4_VertexVisitor(const e4_Node &n);
    e4_VertexVisitor(const e4_Node &n, const char *nm, e4_VertexType ft);

    /*
     * Constructors which initialize the visitor from a vertex.
     */

    e4_VertexVisitor(const e4_Vertex &ff);
    e4_VertexVisitor(const e4_Vertex &ff, bool useVertexName, 
		     bool useVertexType);
    e4_VertexVisitor(const e4_Vertex &ff, bool useVertexName, 
		     bool useVertexType, e4_VisitMethod vmvm);

    /*
     * Constructors that initialize the visitor from a storage.
     */

    e4_VertexVisitor(const e4_Storage &s);
    e4_VertexVisitor(const e4_Storage &s, const char *nm, e4_VertexType ft);

    /*
     * Destructor:
     */

    ~e4_VertexVisitor();

    /*
     * Is this visitor done visiting all the vertices in its itinerary?
     */

    bool IsDone();

    /*
     * Get the next vertex from the visitor.
     */

    bool NextVertex(e4_Vertex &ff);

    /*
     * Get the current vertex.
     */

    bool CurrentVertex(e4_Vertex &ff);

    /*
     * Retrieve the current vertex and advance to the next one.
     */

    bool CurrentVertexAndAdvance(e4_Vertex &ff);

    /*
     * Reset the visitor to a given vertex.
     */

    bool SetVertex(const e4_Vertex &ff);
    bool SetVertex(const e4_Vertex &ff, bool useVertexName, 
		   bool useVertexType);
    bool SetVertex(const e4_Vertex &ff, bool useVertexName, 
		   bool useVertexType, e4_VisitMethod vmvm);

    /*
     * Reset the visitor to start at the first vertex of the given node.
     */

    bool SetNode(const e4_Node &nn);
    bool SetNode(const e4_Node &nn, const char *nm, e4_VertexType ft);

    /*
     * Reset the visitor to visit all vertices in a storage.
     */

    bool SetStorage(const e4_Storage &ss);
    bool SetStorage(const e4_Storage &ss, const char *nm, e4_VertexType ft);

    /*
     * Is this instance valid?
     */

    bool IsValid();

    /*
     * Query methods:
     */

    inline e4_VisitMethod VisitMethod() const {return vm;}
    inline int VisitFilter() const {return vf;}
    inline bool StorageVisited(e4_Storage &ss) const {
	if (s.IsValid()) {
	    ss = s;
	    return true;
	}
	return false;
    }
    inline bool VertexVisited(e4_Vertex &ff) const {
	if (s.IsValid() && f.IsValid()) {
	    ff = f;
	    return true;
	}
	return false;
    }
    inline e4_VertexType TypeFilter() const {return typeID;}
    const char *NameFilter() const;
    bool NodeVisited(e4_Node &n) const;
};

/*
 * class e4_MarkerVisitor:
 */

class e4_DLL e4_MarkerVisitor {
private:
    /*
     * The storage used by this visitor.
     */

    e4_Storage s;

    /*
     * The last marker returned by the visitor.
     */

    e4_Marker t;

    /*
     * Is this visitor done visiting all the markers in this storage?
     */

    bool done;

public:

    /*
     * Default constructor:
     */

    e4_MarkerVisitor();

    /*
     * Copying constructor:
     */

    e4_MarkerVisitor(const e4_MarkerVisitor &referrer);

    /*
     * Assignment operator:
     */

    e4_MarkerVisitor & operator=(const e4_MarkerVisitor &referrer);

    /*
     * Comparison operators:
     */

    bool operator==(const e4_MarkerVisitor &compared) const;
    bool operator!=(const e4_MarkerVisitor &compared) const;

    /*
     * Constructors that initialize the visitor from a storage, marker, node
     * or vertex.
     */

    e4_MarkerVisitor(const e4_Storage &ss);
    e4_MarkerVisitor(const e4_Marker &tt);
    e4_MarkerVisitor(const e4_Node &nn);
    e4_MarkerVisitor(const e4_Vertex &ff);

    /*
     * Destructor:
     */

    ~e4_MarkerVisitor();

    /*
     * Is this visitor done visiting all the markers in this storage?
     */

    bool IsDone();

    /*
     * Get the next marker from the visitor:
     */

    bool NextMarker(e4_Marker &tt);

    /*
     * Get the current marker.
     */

    bool CurrentMarker(e4_Marker &tt);

    /*
     * Get the current marker and advance to the next one.
     */

    bool CurrentMarkerAndAdvance(e4_Marker &tt);

    /*
     * Reset the visitor from a given storage, marker, node or vertex.
     */

    bool SetStorage(const e4_Storage &ss);
    bool SetMarker(const e4_Marker &tt);
    bool SetNode(const e4_Node &nn);
    bool SetVertex(const e4_Vertex &ff);

    /*
     * Is this visitor valid?
     */

    bool IsValid();
};

/*
 * class e4_NodeVisitor:
 */

class e4_DLL e4_NodeVisitor {
private:
    /*
     * The storage used by this visitor.
     */

    e4_Storage s;

    /*
     * The last node returned by the visitor.
     */

    e4_Node n;

    /*
     * Is this visitor done visiting all the nodes in this storage?
     */

    bool done;

public:

    /*
     * Default constructor:
     */

    e4_NodeVisitor();

    /*
     * Copying constructor:
     */

    e4_NodeVisitor(const e4_NodeVisitor &referrer);

    /*
     * Assignment operator:
     */

    e4_NodeVisitor & operator=(const e4_NodeVisitor &referrer);

    /*
     * Comparison operators:
     */

    bool operator==(const e4_NodeVisitor &compared) const;
    bool operator!=(const e4_NodeVisitor &compared) const;

    /*
     * Constructors that initialize the visitor from a storage, marker, node
     * or vertex.
     */

    e4_NodeVisitor(const e4_Storage &ss);
    e4_NodeVisitor(const e4_Marker &tt);
    e4_NodeVisitor(const e4_Node &nn);
    e4_NodeVisitor(const e4_Vertex &ff);

    /*
     * Destructor:
     */

    ~e4_NodeVisitor();

    /*
     * Is this visitor done visiting all the nodes in this storage?
     */

    bool IsDone();

    /*
     * Get the next node from this visitor:
     */

    bool NextNode(e4_Node &nn);

    /*
     * Get the current node.
     */

    bool CurrentNode(e4_Node &nn);

    /*
     * Get the current node and advance the visitor to the next node.
     */

    bool CurrentNodeAndAdvance(e4_Node &nn);

    /*
     * Reset the visitor from a given storage, marker, node or vertex.
     */

    bool SetStorage(const e4_Storage &ss);
    bool SetMarker(const e4_Marker &tt);
    bool SetNode(const e4_Node &nn);
    bool SetVertex(const e4_Vertex &ff);

    /*
     * Is this visitor valid?
     */

    bool IsValid();
};

/*
 * A way to transmit arbitrary values:
 */

typedef struct e4_Value {
    e4_VertexType vertexType;	/* Type of the value being passed. */
    e4_Node n;
    union {
	int i;
        double d;
        char *s;
        e4_Binary b;
    } u;
} e4_Value;

/*
 * Global constant instances of e4_RefCount, e4_Storage, e4_Marker, e4_Node
 * and e4_Vertex that are guaranteed to be invalid. You can assign these
 * to a variable of the appropriate type to ensure that the reference
 * previously stored in that variable is discarded. Especially useful for
 * removing remaining references in heap-allocated structures.
 */

extern e4_DLL const e4_RefCount	invalidRefCount;
extern e4_DLL const e4_Storage	invalidStorage;
extern e4_DLL const e4_Marker	invalidMarker;
extern e4_DLL const e4_Node	invalidNode;
extern e4_DLL const e4_Vertex	invalidVertex;

#include "e4id.h"

#endif	/* __E4_GRAPH_H__ */
