/*
 * e4hash.h --
 *
 * Declarations for the e4_HashTable and attendant structures.
 *
 * This file was extracted verbatim from generic/tcl.h in the Tcl 8.0.4
 * distribution. The names of every structure and function was changed to
 * have the e4_ prefix, to prevent name conflicts when the e4tree library 
 * is linked with Tcl.
 *
 * Copyright (c) 1991-1993 The Regents of the University of California.
 * Copyright (c) 1994 Sun Microsystems, Inc.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 *	Authors: Jacob Levy and Jean-Claude Wippler.
 *		 jyl@best.com	jcw@equi4.com
 *
 *	Copyright: JYL Software, Inc., (c) 1999.
 */

#ifndef	__E4_HASH_H__
#define __E4_HASH_H__

#ifdef	__cplusplus
extern "C" {			/* Entire file! */
#endif

#include <stdio.h>
#include <string.h>

#if defined (_MSC_VER_)
#include <malloc.h>
#else
#include <stdlib.h>
#endif

/*
 * Include the definition of e4_DLL:
 */

#include "e4dlldef.h"

/*
 * Forward declaration of E4_HashTable.  Needed by some C++ compilers
 * to prevent errors when the forward reference to E4_HashTable is
 * encountered in the E4_HashEntry structure.
 */

#ifdef __cplusplus
struct e4_HashTable;
#endif

/*
 * Structure definition for an entry in a hash table.
 */

typedef struct e4_HashEntry {
    struct e4_HashEntry *nextPtr;	/* Pointer to next entry in this
					 * hash bucket, or NULL for end of
					 * chain. */
    struct e4_HashTable *tablePtr;	/* Ptr to table containing entry. */
    struct e4_HashEntry **bucketPtr;	/* Pointer to bucket that points to
					 * first entry in this entry's chain:
					 * used for deleting the entry. */
    void *clientData;			/* Application stores something here
					 * with e4_SetHashValue. */
    union {				/* Key has one of these forms: */
	char *oneWordValue;		/* One-word value for key. */
	int words[1];			/* Multiple integer words for key.
					 * The actual size will be as large
					 * as necessary for this table's
					 * keys. */
	char string[4];			/* String for key.  The actual size
					 * will be as large as needed to hold
					 * the key. */
    } key;				/* MUST BE LAST FIELD IN RECORD!! */
} e4_HashEntry;

/*
 * Structure definition for a hash table.  Must be in e4.h so clients
 * can allocate space for these structures, but clients should never
 * access any fields in this structure.
 */

#define E4_SMALL_HASH_TABLE 4
typedef struct e4_HashTable {
    e4_HashEntry **buckets;		/* Pointer to bucket array.  Each
					 * element points to first entry in
					 * bucket's hash chain, or NULL. */
    e4_HashEntry *staticBuckets[E4_SMALL_HASH_TABLE];
					/* Bucket array used for small tables
					 * (to avoid mallocs and frees). */
    int numBuckets;			/* Total number of buckets allocated
					 * at **bucketPtr. */
    int numEntries;			/* Total number of entries present
					 * in table. */
    int rebuildSize;			/* Enlarge table when numEntries gets
					 * to be this large. */
    int downShift;			/* Shift count used in hashing
					 * function.  Designed to use high-
					 * order bits of randomized keys. */
    int mask;				/* Mask value used in hashing
					 * function. */
    int keyType;			/* Type of keys used in this table. 
					 * It's either E4_STRING_KEYS,
					 * E4_ONE_WORD_KEYS, or an integer
					 * giving the number of ints that
                                         * is the size of the key.
					 */
    e4_HashEntry *(*findProc) (struct e4_HashTable *tablePtr,
			       const char *key);
    e4_HashEntry *(*createProc) (struct e4_HashTable *tablePtr,
				 const char *key, int *newPtr);
} e4_HashTable;

/*
 * Structure definition for information used to keep track of searches
 * through hash tables:
 */

typedef struct e4_HashSearch {
    e4_HashTable *tablePtr;		/* Table being searched. */
    int nextIndex;			/* Index of next bucket to be
					 * enumerated after present one. */
    e4_HashEntry *nextEntryPtr;	/* Next entry to be enumerated in the
					 * the current bucket. */
} e4_HashSearch;

/*
 * Acceptable key types for hash tables:
 */

#define E4_STRING_KEYS		0
#define E4_ONE_WORD_KEYS	1

/*
 * Macros for clients to use to access fields of hash entries:
 */

#define E4_GETHASHVALUE(h) ((h)->clientData)
#define E4_SETHASHVALUE(h, value) ((h)->clientData = (void *) (value))
#define E4_GETHASHKEY(tablePtr, h) \
    ((char *) (((tablePtr)->keyType == E4_ONE_WORD_KEYS) \
	       ? (h)->key.oneWordValue \
	       : (h)->key.string))

/*
 * Macros to use for clients to use to invoke find and create procedures
 * for hash tables:
 */

#define E4_FINDHASHENTRY(tablePtr, key) \
	(*((tablePtr)->findProc))(tablePtr, key)
#define E4_CREATEHASHENTRY(tablePtr, key, newPtr) \
	(*((tablePtr)->createProc))(tablePtr, key, newPtr)

/*
 * Function entry points:
 */

extern e4_DLL void		e4_DeleteHashEntry (e4_HashEntry *entryPtr);
extern e4_DLL void		e4_DeleteHashTable (e4_HashTable *tablePtr);
extern e4_DLL e4_HashEntry *	e4_FirstHashEntry (e4_HashTable *tablePtr,
						   e4_HashSearch *searchPtr);
extern e4_DLL e4_HashEntry *	e4_NextHashEntry (e4_HashSearch *searchPtr);
extern e4_DLL e4_HashTable *	e4_NewHashTable (int keyType);

#ifdef	__cplusplus
}					/* End of extern "C". */
#endif	/* __cplusplus */

#endif	/* __E4_HASH_H__ */
