/*
 * e4id.h --
 *
 *	This file contains the definition of unique ID wrapper types. We
 *	provide these types to ensure type-correct use of unique IDs. The
 *	following classes are provided:
 *
 *	e4_CommonUniqueID	-- Defines the common parts for all wrappers.
 *	e4_MarkerUniqueID	-- A wrapper for a unique ID for a marker.
 *	e4_NodeUniqueID		-- A wrapper for a unique ID for a node.
 *	e4_VertexUniqueID	-- A wrapper for a unique ID for a vertex.
 *
 * Copyright (c) 2000-2001, JYL Software Inc.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE, EVEN IF
 * JYL SOFTWARE INC. IS MADE AWARE OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef	__E4_ID_H__
#define __E4_ID_H__

/*
 * The parts common to all wrappers:
 */

class e4_DLL e4_CommonUniqueID {
private:
    /*
     * The wrapped ID and SP values:
     */

    int ID;
    int SP;

protected:
    /*
     * Get and assign the ID, SP fields.
     */

    inline int GetID() const {return ID;}
    inline void SetID(int id) {ID = id;}

    inline int GetSP() const {return SP;}
    inline void SetSP(int sp) {SP = sp;}

public:
    /*
     * Default constructor:
     */

    e4_CommonUniqueID();

    /*
     * Copying constructor:
     */

    e4_CommonUniqueID(const e4_CommonUniqueID &referrer);

    /*
     * Constructor that assigns a unique ID to the wrapper:
     */

    e4_CommonUniqueID(int uid, int sp);

    /*
     * Assignment operator:
     */

    e4_CommonUniqueID & operator=(const e4_CommonUniqueID &referrer);

    /*
     * Destructor:
     */

    ~e4_CommonUniqueID();

    /*
     * Comparison operators:
     */

    bool operator==(const e4_CommonUniqueID &compared) const;
    bool operator!=(const e4_CommonUniqueID &compared) const;

    /*
     * Is this unique ID wrapper assigned a value?
     */

    inline bool IsAssigned() const 
    {
	return ((ID == E4_INVALIDUNIQUEID) || (SP == 0))? false : true;
    }
};

/*
 * Wrapper for unique ID derived from an e4_Marker.
 */

class e4_DLL e4_MarkerUniqueID : public e4_CommonUniqueID {
public:
    /*
     * Default constructor:
     */

    e4_MarkerUniqueID();

    /*
     * Copying constructor:
     */

    e4_MarkerUniqueID(const e4_MarkerUniqueID &referrer);

    /*
     * Constructor that assigns a value.
     */

    e4_MarkerUniqueID(int mid, int sp);

    /*
     * Assignment operator:
     */

    e4_MarkerUniqueID & operator=(const e4_MarkerUniqueID &referrer);

    /*
     * Comparison operators:
     */

    bool operator==(const e4_MarkerUniqueID &compared) const;
    bool operator!=(const e4_MarkerUniqueID &compared) const;

    /*
     * Extract the value.
     */

    inline operator int() const {return GetID();}

    /*
     * Extract the value.
     */

    inline int GetUniqueID() const {return GetID();}

    /*
     * Set the value.
     */

    inline void SetUniqueID(int nid, e4_Storage s) {
	if (!s.IsValid()) {
	    return;
	}
	SetID(nid);
	SetSP(s.GetTemporaryUID());
    }
};

/*
 * Wrapper for unique ID derived from an e4_Node.
 */

class e4_DLL e4_NodeUniqueID : public e4_CommonUniqueID {
public:
    /*
     * Default constructor:
     */

    e4_NodeUniqueID();

    /*
     * Copying constructor:
     */

    e4_NodeUniqueID(const e4_NodeUniqueID &referrer);

    /*
     * Constructor that assigns a value.
     */

    e4_NodeUniqueID(int nid, int sp);

    /*
     * Assignment operator:
     */

    e4_NodeUniqueID & operator=(const e4_NodeUniqueID &referrer);

    /*
     * Comparison operators:
     */

    bool operator==(const e4_NodeUniqueID &compared) const;
    bool operator!=(const e4_NodeUniqueID &compared) const;

    /*
     * Extract the value.
     */

    inline operator int() const {return GetID();}

    /*
     * Extract the value.
     */

    inline int GetUniqueID() const {return GetID();}

    /*
     * Set the value.
     */

    inline void SetUniqueID(int nid, e4_Storage s) {
	if (!s.IsValid()) {
	    return;
	}
	SetID(nid);
	SetSP(s.GetTemporaryUID());
    }
};

/*
 * Wrapper for unique ID derived from an e4_Vertex.
 */

class e4_DLL e4_VertexUniqueID : public e4_CommonUniqueID {
public:
    /*
     * Default constructor:
     */

    e4_VertexUniqueID();

    /*
     * Copying constructor:
     */

    e4_VertexUniqueID(const e4_VertexUniqueID &referrer);

    /*
     * Constructor that assigns a value.
     */

    e4_VertexUniqueID(int vid, int sp);

    /*
     * Assignment operator:
     */

    e4_VertexUniqueID & operator=(const e4_VertexUniqueID &referrer);

    /*
     * Comparison operators:
     */

    bool operator==(const e4_VertexUniqueID &compared) const;
    bool operator!=(const e4_VertexUniqueID &compared) const;

    /*
     * Extract the value.
     */

    inline operator int() const {return GetID();}

    /*
     * Extract the value.
     */

    inline int GetUniqueID() const {return GetID();}

    /*
     * Set the value.
     */

    inline void SetUniqueID(int nid, e4_Storage s) {
	if (!s.IsValid()) {
	    return;
	}
	SetID(nid);
	SetSP(s.GetTemporaryUID());
    }
};

#endif	/* __E4_ID_H__ */

