/*
 * e4xml.h --
 *
 *	Definition of the e4_XML classes.
 *
 *	The e4_XMLParser class wraps around James Clark's expat
 *	parser and stores the parse tree under a given node.
 *
 *	The e4_XMLGenerator class generates XML output from a given
 *	node on the supplied FILE * (Not yet implemented).
 *
 *	Authors: Jacob Levy and Jean-Claude Wippler.
 *		 jyl@best.com	jcw@equi4.com
 *
 *	Copyright: JYL Software, Inc., (c) 2000-2001.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE, EVEN IF
 * JYL SOFTWARE INC. IS MADE AWARE OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef	__E4_XML_H__
#define	__E4_XML_H__

#include <stdio.h>
#include <stdlib.h>

#include "expat.h"
#include "e4graph.h"
#include "e4xmldll.h"
#include "e4base64.h"

class e4_XMLDLL e4_XMLParser {
private:

    /*
     * Is this parser ready?
     */

    bool ready;

    /*
     * Does this parser have an error flagged?
     */

    bool error;

    /*
     * If so, this is the error message.
     */

    char *errorString;

    /*
     * Has the parse started?
     */

    bool started;

    /*
     * How many open XML elements have we entered?
     */

    int depth;

    /*
     * Are we currently in a vertex-add?
     */

    bool inVertex;

    /*
     * The hash table for storing nodes seen so far during the parse.
     */

    e4_HashTable *nodesSeen;

    /*
     * This is the node under which the XML elements are to be
     * inserted.
     */

    e4_Node n;

    /*
     * This is the storage in which the XML elements are stored.
     */

    e4_Storage s;

    /*
     * This is the XML parser itself.
     */

    void *parser;

    /*
     * BASE64 string.
     */

    byte *base64bytes;

    /*
     * Dynamic string containing character data as it is being read in.
     */

    e4_DString *dsp;

public:

    /*
     * This constructor constructs an empty parser:
     */

    e4_XMLParser();

    /*
     * This constructor sets up the parser to store new elements under 'nn'.
     */

    e4_XMLParser(e4_Node nn);

    /*
     * The destructor cleans up the state associated with this parser.
     */

    virtual ~e4_XMLParser();

    /*
     * This operation assigns a node as the current node of the parser.
     * New elements will be stored under this node.
     */

    void SetNode(e4_Node nn);

    /*
     * Hash a node so backwards references to it can be reconstructed.
     */

    void HashNode(e4_Node nn, int id) const;

    /*
     * This operation returns in 'nn' the current node.
     */

    bool GetNode(e4_Node &nn);

    /*
     * ClearError clears a stored error state.
     */

    void ClearError();

    /*
     * ErrorString returns the error string for the current error state.
     */

    const char *ErrorString();

    /*
     * Is the parse finished?
     */

    bool Finished();

    /*
     * This operation parses one buffer of XML input and stores the new
     * elements under the stored node. It returns true if all went well,
     * false otherwise. If an error occurred then most likely HasError()
     * will return true and ErrorString() will return a non-NULL string.
     */

    bool Parse(char *buf, size_t len);

    /*
     * Does this parser have an uncleared error state?
     */

    bool HasError();

    /*
     * As a service we export BASE64 decoding functionality. The returned
     * byte buffer must be deleted by the caller when no longer needed.
     */

    byte *Base64_Decode(const char *base64str, int *len);

    /*
     * The operations below must unfortunately be public, even though
     * they are not intended to be used by code outside the implementation
     * of the e4_XMLParser class.
     */

    /*
     * Flag an error on this parser and store an error string.
     */

    void FlagError(const char *msg);

    /*
     * This operation adds a vertex to the current node.
     */

    bool AddVertex(const char *vname, const char *vtype, const char *vval);

    /*
     * Are we currently in a vertex?
     */

    bool InVertex() const;

    /*
     * This operation closes the current vertex.
     */

    void ExitVertex();

    /*
     * This operation adds a back reference to a node that was previously
     * parsed.
     */

    bool AddNodeBackRef(const char *nm, const char *nid);

    /*
     * This operation constructs a new parser from the stored file pointer
     * and node.
     */

    bool ConstructParser();

    /*
     * Increment/decrement the nesting depth of XML elements entered.
     */

    void IncrDepth();
    void DecrDepth();

    /*
     * StoreCharData stores (and auto-initializes) a currently
     * constructed character data string.
     */

    void StoreCharData(char *data, int len);

    /*
     * Ensure that any currently being constructed character data is
     * saved into the persistent storage.
     */

    void FinishCharData();
};

class e4_XMLDLL e4_XMLGenerator {
private:
    /*
     * Is this generator ready to generate?
     */

    bool ready;

    /*
     * The starting node.
     */

    e4_Node sn;

    /*
     * The XML element into which the starting node will be wrapped.
     */

    char *se;

    /*
     * The e4_DString holding the generated XML.
     */

    e4_DString *dsptr;

    /*
     * nodeCounter counts the number of nodes for which we've already generated
     * XML output. Each node is hashed under its unique ID, and the hash value
     * is its assigned nodeCounter. That way we can generate XML for back
     * references to nodes we've already produced XML for before.
     */

    int nodeCounter;

    /*
     * nodesSeen is a hash table for nodes that were already output into the
     * generated XML. The key is the node unique ID, and the value is the
     * assigned nodeCounter value for that node. This value is used in a
     * "__nodebackreference__" element if the same node occurs more than once
     * in the sub-graph from which we are generating the XML output.
     */

    e4_HashTable *nodesSeen;

    /*
     * BASE64 string.
     */

    char *base64str;

public:
    /*
     * Default constructor:
     */

    e4_XMLGenerator();

    /*
     * Constructor that takes a node and element name as input.
     */

    e4_XMLGenerator(e4_Node nn, char *sen);

    /*
     * Destructor:
     */

    virtual ~e4_XMLGenerator();

    /*
     * Set the starting node. Resets the XML output.
     */

    void SetNode(e4_Node nn);

    /*
     * Set the starting element name. Resets the XML output.
     */

    void SetElementName(char *sen);

    /*
     * Set the element name and the starting node. Resets the XML output.
     */

    void SetElementNameAndNode(char *sen, e4_Node nn);

    /*
     * Retrieve the node for which XML output is being generated.
     */

    void GetNode(e4_Node &nn) const;

    /*
     * Retrieve the wrapping XML element tag name.
     */

    char *GetElementName() const;

    /*
     * Produce XML output.
     */

    char *Generate();

    /*
     * Get the accumulated XML output.
     */

    char *Get();

    /*
     * As a service we export BASE64 encoding functionality. The returned
     * character string buffer must be deleted by the caller when no longer
     * needed.
     */

    char *Base64_Encode(byte *bytes, int len);

private:
    /*
     * Reset the state of this generator to empty.
     */

    void Reset();

    /*
     * Generate XML output for a node.
     */

    void GenerateNode(char *elname, e4_Node nn);

    /*
     * Generate XML output for a vertex.
     */

    void GenerateVertex(e4_Vertex vv);

    /*
     * Generate XML output for attributes of a node.
     */

    bool GenerateAttributes(e4_Node attrs) const;

    /*
     * Generate output for DOCTYPE element declarations.
     */

    bool GenerateElementDecls(e4_Node doctype) const;
};

#endif	/* __E4_XML_H__ */
