/*
 * t4graph.h --
 *
 *	Main include file for the T4Graph package.
 *
 *	Authors: Jacob Levy and Jean-Claude Wippler.
 *		 jyl@best.com	jcw@equi4.com
 *
 * Copyright (c) 2000-2001, JYL Software Inc.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE, EVEN IF
 * JYL SOFTWARE INC. IS MADE AWARE OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef	__T4GRAPH_H__
#define	__T4GRAPH_H__

/*
 * External dependencies:
 */

#include "tcl.h"
#include "e4graph.h"
#include "t4dll.h"
#include "t4graph_c.h"

/*
 * Enumeration of the kinds of vertex names. Right now, T4Graph supports
 * two kinds of vertex names:
 *
 * T4VNK_INDEX		"n(3)" or simply "n"
 * T4VNK_RANK		"33" or "-1"
 */

typedef enum T4VertexNameKinds {
    T4VNK_INDEX = 0,
    T4VNK_RANK
} T4VertexNameKinds;

/*
 * Internal include files:
 */

#include "t4graphrep.h"

/*
 * Declarations of all command procedures:
 */

#ifdef	__cplusplus
extern "C" {
#endif

extern int		OpenStorageProc(ClientData cd, Tcl_Interp *interp,
					int objc, Tcl_Obj *CONST objv[]);
extern int		IsOpenStorageProc(ClientData cd, Tcl_Interp *interp,
					  int objc, Tcl_Obj *CONST objv[]);
extern int		TransferStorageProc(ClientData cd, Tcl_Interp *interp,
					    int objc, Tcl_Obj *CONST objv[]);

extern int		StorageCmdProc(ClientData cd, Tcl_Interp *interp,
				       int objc, Tcl_Obj *CONST objv[]);
extern int		MarkerCmdProc(ClientData cd, Tcl_Interp *interp,
				      int objc, Tcl_Obj *CONST objv[]);
extern int		NodeCmdProc(ClientData cd, Tcl_Interp *interp,
				    int objc, Tcl_Obj *CONST objv[]);
extern int		VertexCmdProc(ClientData cd, Tcl_Interp *interp,
				      int objc, Tcl_Obj *CONST objv[]);

/*
 * This procedure cleans up the TGraph related storage when an
 * interpreter is deleted.
 */

extern void		OpenStorageInterpDelProc(ClientData cd,
						 Tcl_Interp *interp);

#ifdef	__cplusplus
}
#endif

/*
 * Helper procedures to manage a per-interpreter registry of open
 * storages.
 */

extern T4Storage *	FindRegisteredStorage(Tcl_HashTable *tblPtr,
					      const char *snm,
					      const char *dnm);
extern void		RegisterStorage(Tcl_HashTable *tblPtr,
					T4Storage *s,
					const char *cmdname,
					const char *snm, const char *dnm);
extern void		UnregisterStorage(Tcl_HashTable *tblPtr,
					  T4Storage *s,
					  const char *cmdname,
					  const char *snm,
					  const char *dnm);

/*
 * Create Tcl commands to manipulate a specific T4Graph object:
 */

extern int		MakeStorageCommand(Tcl_Interp *interp, T4Storage *s);
extern char *		MakeMarkerCommand(Tcl_Interp *interp, T4Marker *mp,
					  e4_MarkerUniqueID muid);
extern char *		MakeNodeCommand(Tcl_Interp *interp, T4Node *np,
					e4_NodeUniqueID nuid);
extern char *		MakeVertexCommand(Tcl_Interp *interp, T4Vertex *vp,
					  e4_VertexUniqueID vuid);

/*
 * Delete Tcl commands that were created to manipulate T4Graph objects:
 */

extern void		DeleteStorageCommand(Tcl_Interp *interp, T4Storage *s);
extern void		DeleteMarkerCommand(Tcl_Interp *interp, T4Marker *t);
extern void		DeleteNodeCommand(Tcl_Interp *interp, T4Node *n);
extern void		DeleteVertexCommand(Tcl_Interp *interp, T4Vertex *f);

/*
 * Dispatch to a type-specific object procedure according to the
 * sub-object obtained from the "-with subobj" specification.
 */

extern int		DispatchSubObject(Tcl_Interp *interp,
					  T4Storage *s,
					  T4InternalRep *repPtr,
					  int objc,
					  Tcl_Obj *CONST objv[]);

/*
 * Manage stored procedures and their definition and invocation in Tcl.
 */

extern int		DefineStoredProcedure(Tcl_Interp *interp,
					      char *cmdname,
					      char *nodename,
					      const char *storedproc);
extern int		InvokeStoredProcedure(Tcl_Interp *interp,
					      char *cmdname,
					      int objc,
					      Tcl_Obj *CONST objv[]);

/*
 * Parse a vertex specification. The vertex name can be one of the
 * allowed formats as defined by enum T4VertexNameKinds.
 */

extern int		ParseVertexName(Tcl_Interp *interp, char *spec,
					char **fnp, int *indexp,
					T4VertexNameKinds *vnkp);

/*
 * Parse an insert order name. The name is a string. It is translated
 * by this function into one of the values acceptable for e4_InsertOrder.
 */

extern int		ParseInsertOrder(Tcl_Interp *interp, Tcl_Obj *spec,
					 e4_InsertOrder *iop);

/*
 * The names of Tcl procedures that are called in a master to
 * allow/disallow specific operations for safe slave interpreters.
 * Each procedure either succeeds, to indicate the operation is
 * allowed, or fails to indicate that it is disallowed.
 */

#define ALLOW_OPEN_CMD		"tgraph::allow_open"
#define ALLOW_MOD_CMD		"tgraph::allow_mod"
#define ALLOW_ADD_CMD		"tgraph::allow_add"
#define ALLOW_DEL_CMD		"tgraph::allow_del"
#define ALLOW_MOVE_CMD		"tgraph::allow_move"
#define ALLOW_RENAME_CMD	"tgraph::allow_rename"

#endif	/* __T4GRAPH_H__ */
