/*
 * t4graphrep --
 *
 *	This include file provides definitions for the internal representation
 *	of T4Graph objects.
 *
 *	Authors: Jacob Levy and Jean-Claude Wippler.
 *		 jyl@best.com	jcw@equi4.com
 *
 * Copyright (c) 2000-2001, JYL Software Inc.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE, EVEN IF
 * JYL SOFTWARE INC. IS MADE AWARE OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef	__T4GRAPHREP_H__
#define	__T4GRAPHREP_H__

#include "t4dll.h"

/*
 * Forward declarations of the internal representation types, so they
 * can refer to each other.
 */

class e4_TGRAPHDLL T4Storage;
class e4_TGRAPHDLL T4Marker;
class e4_TGRAPHDLL T4Node;
class e4_TGRAPHDLL T4Vertex;

/*
 * The key for use of our private interpreter associated data.
 */

#define T4_ASSOCKEY	"tgraph::openstorages"

/*
 * Enumeration of the different T4Graph internal representations:
 */

typedef enum T4Kinds {
    T4GRAPH_ILLEGAL = -1,
    T4GRAPH_STORAGE = 0,
    T4GRAPH_MARKER,
    T4GRAPH_NODE,
    T4GRAPH_VERTEX
} T4Kinds;

/*
 * This record contains information pertaining to one callback declaration.
 */

typedef struct T4CallbackRecord {
    T4Storage *storage;		/* The storage on which the callback occurs. */
    Tcl_Interp *interp;		/* The interp in which the callback occurs. */
    e4_CallbackKind kind;	/* The kind of callback this record is for. */
} T4CallbackRecord;

/*
 * Enumeration to allow selection of an object kind (for Get and Callback
 * operations).
 */

typedef enum T4ObjectKindSelector {
    T4_OKMARKER = 0,
    T4_OKNODE = 1,
    T4_OKVERTEX = 2
} T4ObjectKindSelector;

/*
 * Enumerations for managing the callback facility on a T4Storage object.
 */

typedef enum T4CallbackEventSelector {
    T4_CBESADD = 0,
    T4_CBESDEL = 1,
    T4_CBESMOD = 2
} T4CallbackEventSelector;

typedef enum T4CallbackActionSelector {
    T4_CBASADD = 0,
    T4_CBASDEL = 1,
    T4_CBASGET = 2,
    T4_CBASSET = 3,
    T4_CBASKND = 4,
    T4_CBASCNT = 5,
    T4_CBASHAS = 6
} T4CallbackActionSelector;

/*
 * Flags marking a T4Vertex.
 */

#define T4_VERTEXEXPORTED	(1 << 0)
#define T4_VERTEXUSED		(1 << 1)

/*
 * How many vertices can be cached between subsequent sweeps of the cache?
 */

#define T4_SWEEP_INTERVAL	256

/*
 * This class serves as the common base class for all T4Graph internal
 * representations, so its possible to pass any of them throuh a
 * common procedure argument.
 */

class e4_TGRAPHDLL T4InternalRep
{
  private:

    /*
     * The name assigned to this T4Graph object in Tcl and its length.
     */

    char *nm;
    int nmlen;

  public:

    /*
     * Default constructor:
     */

    T4InternalRep();

    /*
     * Return a T4Kinds identifier for the type of this
     * T4Graph object.
     */

    virtual T4Kinds KindIdentifier() const;

    /*
     * Return the name by which this object is known in Tcl.
     */

    char *GetName() const;

    /*
     * Return the length of the name by which this object is known in Tcl.
     */

    int GetNameLen() const;

    /*
     * Set the name by which this object is known in Tcl.
     */

    void SetName(char *name);
};

/*
 * This class is the internal representation of a storage object:
 */

class e4_TGRAPHDLL T4Storage : public T4InternalRep
{
  private:
    /*
     * The actual storage object.
     */

    e4_Storage s;

    /*
     * These hash tables contain the internal representations of
     * all objects in this storage that have been exported to Tcl.
     */

    Tcl_HashTable *exportedMarkers;
    Tcl_HashTable *exportedNodes;
    Tcl_HashTable *exportedVertices;

    /*
     * Hash table for callbacks.
     */

    Tcl_HashTable *callbacks;

    /*
     * Counters for how many callbacks there are for each kind of event.
     * When a counter is zero, the C++ level callback is cancelled.
     */

    int cbaddmarker;
    int cbdelmarker;
    int cbaddnode;
    int cbdelnode;
    int cbaddvertex;
    int cbdelvertex;
    int cbmodvertex;

    /*
     * Callback tokens to be used when the callback is deleted.
     */

    T4CallbackRecord *cbAddMarkerRecord;
    T4CallbackRecord *cbDelMarkerRecord;
    T4CallbackRecord *cbAddNodeRecord;
    T4CallbackRecord *cbDelNodeRecord;
    T4CallbackRecord *cbAddVertexRecord;
    T4CallbackRecord *cbDelVertexRecord;
    T4CallbackRecord *cbModVertexRecord;

    /*
     * Interpreter in which this storage is accessible.
     */

    Tcl_Interp *associatedInterp;

    /*
     * The prefix to add to each name exported to Tcl to ensure that
     * the names are unique.
     */

    char *prefix;
    int prefixlen;

    /*
     * How many vertices were cached since the last sweep of the stored
     * vertex table?
     */

    int verticesCachedSinceLastSweep;

    /*
     * Visitor methods:
     */

    int VisitMarkers(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
    int VisitNodes(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
    int VisitVertices(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);

    /*
     * Helper methods for callback registration mgmt:
     */

    int CBAddCallback(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
    int CBDelCallback(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
    int CBGetCallback(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
    int CBSetCallback(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
    int CBKindCallback(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
    int CBCountCallback(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
    int CBHasCallback(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);

    /*
     * Clean up in-core state associated with this storage.
     */

    int InternalClose(Tcl_Interp *interp);

  public:

    /*
     * Constructor:
     */

    T4Storage(e4_Storage ss, char *name);

    /*
     * Destructor:
     */

    virtual ~T4Storage();

    /*
     * This method returns the kind of internal representation that
     * is implemented by this class.
     */

    virtual inline T4Kinds KindIdentifier() const {
	return T4GRAPH_STORAGE;
    }

    /*
     * The following methods implement Tcl sub-commands:
     */

    int AutoCommit(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
    int Close(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
    int Commit(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
    int IsDirty(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
    int MarkDirty(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
    int CopyTo(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
    int Delete(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
    int MarkerExists(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
    int MarkerCount(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
    int Marker(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
    int MarkedNode(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
    int IsEmpty(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
    int MakeEmpty(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
    int IsValid(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
    int Name(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
    int Driver(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
    int Foreach(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
    int Statistic(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
    int Callback(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
    int Get(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);

    /*
     * These operations return internal representations of T4Graph
     * objects identified by index or name. The object must already have
     * been created.
     */

    T4Marker *GetMarkerById(e4_MarkerUniqueID muid);
    T4Node *GetNodeById(e4_NodeUniqueID nuid);
    T4Vertex *GetVertexById(e4_VertexUniqueID vuid);

    T4Marker *GetMarkerByName(char *n);
    T4Node *GetNodeByName(char *n);
    T4Vertex *GetVertexByName(char *n);

    /*
     * These operations create a new entry in the respective hash
     * table when an object is exported to Tcl.
     */

    void StoreMarker(T4Marker *t, int id);
    void StoreNode(T4Node *n, int id);
    void StoreVertex(T4Vertex *f, int id);

    /*
     * Remove unused vertices from the stored vertex table.
     */

    void CleanupStoredVertices();

    /*
     * These operations remove an internal representation from the
     * respective hash table when the object is no longer being
     * exported to Tcl.
     */

    void RemoveMarker(e4_MarkerUniqueID muid);
    void RemoveNode(e4_NodeUniqueID nuid);
    void RemoveVertex(e4_VertexUniqueID vuid);

    /*
     * This class method returns the next name available for export to Tcl
     * for a T4Graph storage object.
     */

    static char *NextStorageName();

    /*
     * This operation removes all T4Graph objects exported to the
     * given interpreter:
     */

    void RemoveAllExportedObjects(Tcl_Interp *interp);

    /*
     * This operation removes all registered callbacks for this storage.
     */

    void RemoveAllCallbacks();

    /*
     * This operation externalizes the e4Graph storage object.
     */

    void ExternalizeStorage(e4_Storage &ss);

    /*
     * These methods get called when a callback event occurs.
     */

    void AddMarkerCallback(Tcl_Interp *interp, e4_Marker m);
    void DelMarkerCallback(Tcl_Interp *interp, e4_Marker m);
    void AddNodeCallback(Tcl_Interp *interp, e4_Node n);
    void DelNodeCallback(Tcl_Interp *interp, e4_Node n);
    void AddVertexCallback(Tcl_Interp *interp, e4_Vertex v);
    void DelVertexCallback(Tcl_Interp *interp, e4_Vertex v);
    void ModVertexCallback(Tcl_Interp *interp, e4_Vertex v);

    /*
     * Associate an interpreter with this storage and retrieve the
     * associated interpreter.
     */

    void SetAssociatedInterpreter(Tcl_Interp *interp);
    Tcl_Interp *GetAssociatedInterpreter();

    /*
     * These methods are called for automatic callbacks when a storage
     * element is deleted.
     */

    void CleanupMarker(e4_Marker m);
    void CleanupNode(e4_Node n);
    void CleanupVertex(e4_Vertex v);
};

/*
 * This class is the internal representation of a marker object.
 */

class e4_TGRAPHDLL T4Marker : public T4InternalRep
{
  private:

    /*
     * The actual marker object.
     */

    e4_Marker t;

    /*
     * The T4Storage instance for the storage containing this marker.
     */

    T4Storage *s;

  public:

    /*
     * Constructor:
     */

    T4Marker(e4_Marker tt, char *name, T4Storage *ss);

    /*
     * Destructor:
     */

    virtual ~T4Marker();

    /*
     * This method returns the kind of internal representation that
     * is implemented by this class.
     */

    virtual inline T4Kinds KindIdentifier() const {
	return T4GRAPH_MARKER;
    }

    /*
     * Retrieve the T4Storage instance for the storage containing
     * this vertex:
     */

    inline T4Storage *GetStorage() const {
	return s;
    }

    /*
     * The following methods implement Tcl sub-commands:
     */

    int Name(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
    int MarkedNode(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
    int SetMarkedNode(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
    int Delete(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
    int Dispose(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
    int IsValid(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
    int Id(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
    int UserData(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);

    /*
     * This operation externalizes the e4Graph marker object.
     */

    void ExternalizeMarker(e4_Marker &tt);
};

/*
 * This class is the internal representation of a node object.
 */

class e4_TGRAPHDLL T4Node : public T4InternalRep
{
  private:

    /*
     * The actual node object.
     */

    e4_Node n;

    /*
     * The T4Storage instance of the storage containing this vertex:
     */

    T4Storage *s;

    /*
     * Cache of vertices that are exported to Tcl.
     */

    Tcl_HashTable *vertexcache;

    /*
     * Obtain a T4Vertex given a vertex name.
     */

    T4Vertex *ObtainVertex(Tcl_Interp *interp, char *vn);

    /*
     * Maybe create a named vertex.
     */

    T4Vertex *MaybeCreateVertex(Tcl_Interp *interp, char *vn);

    /*
     * Private visitor functions:
     */

    int VisitVertices(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
    int VisitMarkers(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
    int VisitParents(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);

  public:

    /*
     * Constructor:
     */

    T4Node(e4_Node nn, char *name, T4Storage *ss);

    /*
     * Destructor
     */

    virtual ~T4Node();

    /*
     * Clear the vertex cache.
     */

    void ClearVertexCache();

    /*
     * Remove one vertex from the vertex cache.
     */

    void RemoveVertexFromCache(T4Vertex *vp);

    /*
     * This method returns the kind of internal representation that
     * is implemented by this class.
     */

    virtual inline T4Kinds KindIdentifier() const {
	return T4GRAPH_NODE;
    }

    /*
     * Retrieve the T4Storage instance for the storage containing
     * this vertex:
     */

    inline T4Storage *GetStorage() const {
	return s;
    }

    /*
     * The following methods implement Tcl sub-commands.
     */

    int VertexCount(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
    int Set(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
    int Add(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
    int Get(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
    int SetNode(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
    int AddNode(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
    int GetVertex(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
    int MoveVertex(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
    int DeleteVertex(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
    int VertexType(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
    int VertexRank(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
    int VertexName(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
    int RenameVertex(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
    int Exists(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
    int Parent(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
    int ParentCount(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
    int OccurrenceCount(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
    int ParentRank(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
    int IsRoot(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
    int RankInParent(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
    int NameInParent(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
    int AddMarker(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
    int MarkWith(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
    int Unmark(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
    int MarkerCount(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
    int IsMarked(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
    int Marker(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
    int Delete(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
    int Dispose(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
    int IsValid(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
    int Method(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
    int Call(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
    int Foreach(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
    int Id(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
    int UserData(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
    int DeleteFirstVertexWithNode(Tcl_Interp *interp, int objc,
				  Tcl_Obj *CONST objv[]);

    /*
     * This operation externalizes the e4Graph node object.
     */

    void ExternalizeNode(e4_Node &nn);
};

/*
 * This class is the internal representation of a vertex object.
 */

class e4_TGRAPHDLL T4Vertex : public T4InternalRep
{
  private:

    /*
     * Flags marking this object.
     */

    int flags;

    /*
     * The actual vertex object.
     */

    e4_Vertex f;

    /*
     * The T4Storage instance of the storage containing this vertex:
     */

    T4Storage *s;

    /*
     * The command info for the stored procedure for this vertex. Null if
     * the procedure has not yet been exported to Tcl.
     */

    Tcl_CmdInfo *storedcmdinfo;

    /*
     * Cached object value for the value of this vertex.
     */

    Tcl_Obj *storedobj;

  public:

    /*
     * Constructor:
     */

    T4Vertex(e4_Vertex ff, char *name, T4Storage *ss);

    /*
     * Destructor:
     */

    virtual ~T4Vertex();

    /*
     * Clear the vertex cache of the node containing this vertex.
     */

    void NodeClearVertexCache();

    /*
     * Remove this vertex from the cache in the node that contains
     * this vertex.
     */

    void NodeRemoveVertexFromCache();

    /*
     * This method returns the kind of internal representation that
     * is implemented by this class.
     */

    virtual inline T4Kinds KindIdentifier() const {
	return T4GRAPH_VERTEX;
    }

    /*
     * Retrieve the T4Storage instance for the storage containing
     * this vertex:
     */

    inline T4Storage *GetStorage() const {
	return s;
    }

    /*
     * Set or unset a flag.
     */

    inline void SetFlag(int f) {
	flags |= f;
    }
    inline void UnsetFlag(int f) {
	flags &= ~(f);
    }
    inline bool IsFlagSet(int f) {
	return ((flags & f) == f) ? true : false;
    }

    /*
     * The following operation is the same as Set() except that it
     * converts the given value to a requested type.
     */

    int SetAs(Tcl_Interp *interp, Tcl_Obj *val, int reqtype);

    /*  
     * The following methods implement Tcl sub-commands.
     */

    int Get(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
    int Set(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
    int SetNode(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
    int Rank(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
    int Delete(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
    int Dispose(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
    int IsValid(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
    int Type(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
    int Name(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
    int Rename(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
    int Node(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
    int Marker(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
    int Move(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
    int Next(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
    int Prev(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
    int Call(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
    int Id(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
    int UserData(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);

    /*
     * This operation externalizes the e4Graph vertex object.
     */

    void ExternalizeVertex(e4_Vertex &ff);
};

#endif	/* __T4GRAPHREP_H__ */
